; The LAML library and programs written by Kurt Normark, Aalborg University, Denmark.
; Copyright (C) 1999  Kurt Normark.
; 
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


;;;; This is the LAML SchemeDoc tool which extracts interface documentation of a Scheme source program.
;;;; The extracted information can be presented as a web document when used together with the LAML
;;;; <a href = "../../../styles/manual/man/manual.html"> manual document style</a>.<p>
;;;; 
;;;; The LAML SchemeDoc tool is part of the LAML software package, which is free software
;;;; available from the <a href = "http://www.cs.auc.dk/~normark/laml/">LAML home page</a>.
;;;; The current page is created by SchemeDoc.<p>
;;;;
;;;; Interface documentation is found in two, three or four semicolon Lisp comments.
;;;; A number of internal tags in the comments make it possible to express a variety
;;;; of interface properties in a structured way.
;;;; The main procedure of the tool is extract-documentation-from-scheme-file. Two additional functions
;;;; used for internal purposes are also documented.<p>
;;;; 
;;;; If you read this via the documentation and manual facility, this text is extracted
;;;; from the Scheme source file, which implements Schemedoc. As such, Schemedoc has simply
;;;; been used on itself.<p>
;;;; 
;;;; We recommend that you consult <a href="../../../tutorial/schemedoc/schemedoc.html">"SchemeDoc and Manual pages"</a> for a tutorial 
;;;; introduction to SchemeDoc.<p>
;;;;
;;;; Besides SchemeDoc, the LAML software package has a tool for internal documentation of Scheme programs.
;;;; This tool is called the <a href = "../../../styles/elucidator/man/elucidator.html"> Scheme Elucidator</a>.<p>
;;;; 
;;;; The interfaces of all the LAML libraries, tools, and document styles have been documented by use of SchemeDoc.
;;;; As such, SchemeDoc is a well-proven tool - at least relative to the style of Scheme programming the LAML author makes use of.

; Warning: If schemdoc.scm is loaded after manual.scm, and if manual facility is used
; thereafter, problems occur. There is a conflict, yet unknown, between manual.scm and schemedoc.scm

; ---------------------------------------------------------------------------------------------------
; Putting part one and two together:


;; This function is the main function. As such, this function IS the schemedoc tool.
;; The function takes as input a name of a scheme file (full path, file extension inclusive).
;; The function returns a list structures suitable for the manual style to display via the function
;; make-manual. Notice that make-manual is not called from this function.

(define (extract-documentation-from-scheme-file scheme-file)
  (lexical-to-syntactical-comments! scheme-file temp-output-file)
  ; hereby temp-output-file is defined
  (let ((res (extract-relevant-documentation temp-output-file)))
      (if delete-comment-file? (delete-file temp-output-file))
    res))

;; A boolean variable which tells whether to delete the temporary (by this tool) generated source
;; file, in which comments are syntactical Lisp forms. The name of the source file is determined by
;; the variable temp-output-file. By default, the value is #f.
(define delete-comment-file? #f)


;;; How to use comments for documentation purposes. 
;;; In order to have documentation extracted, the documentation is organized in comments
;;; with one, two, three, or four initial semicolons. Recall that semicolons are the Lisp lexical
;;; comment symbols: The rest of a line with a semicolon is a comment.<p>
;;; All one comment semicolon comments are ignored by this facility. N semicolon comments, N higher than one, are called SchemeDoc comments. <p>
;;; Two semicolon comments are used as comments of define forms. You can either place
;;; the comment before the define form or after the name/signature of the define form. <p>
;;; Three semicolon comments
;;; are used for text sections between documented forms.  The first
;;; sentence (separated with '.' and a space) is used as the title of the text section.
;;; (This text your are reading now - in a browser- is written in a three semicolon comment).<p>
;;; Four semicolon comments are used as the introduction to a documentation document. 
;;; There should, as such, only by one four semicolon comment in a Scheme file. <p>
;;; 
;;; Please notice that function definitions of the form <p>
;;; <pre>
;;;   (define f (lambda (par) ...)) <p>
;;; </pre>
;;; should be documented with a two semicolon comment before the form, whereas <p>
;;; <pre>
;;;   (define (f par) ...) <p>
;;; </pre>
;;; can be documented with two semicolon comments before the form or just after (f par). 
;;; If you have two semicolon comments both before and within the define form they are concatenated. <p>
;;; It is possible to use HTML formatting within a schemedoc comment.<p>
;;;
;;; In comments at level two there
;;; may be additional <b>internal tagging</b> at the end of the comment. Here is an example:<p>
;;; 
;;; <pre>;; Return the fifth element of a list
;;; ;; .form (fifth x)
;;; ;; .parameter x must be a list
;;; (define fifth (make-selector-function 5)) 
;;; </pre>
;;; 
;;; The Schemedoc tool parses the comments and returns a list like
;;; 
;;; <pre>
;;; (
;;;  (description "Return the fifth element of a list")
;;;  (form "(fifth x)"
;;;  (param "x must be a list")
;;; )
;;; </pre>
;;; The manual producing tool may decide to parse additional parts of the strings (the cadr parts of the lists shown above).<p>
;;; 
;;; The general format of an internal Schemdoc tag is: <br>
;;;  <kbd>.key description</kbd>.<p>
;;; 
;;; The rules concerning internal tags in Schemedoc comments are the following:
;;; 
;;; <ol>
;;; <li> Each tag starts with a dot '.' and runs through the line, until CR.
;;; <li> The only characters allowed between the semicolons and <kbd>.key</kbd> are blank space characters (space, tab).
;;; <li> Internal tags must be at the end of a comment; no descriptions (besides other internal tags) must be found after them.
;;; <li> key can be an arbitry symbol without spaces. The manual style, however, only recognizes title, form, pre-condition, 
;;;      description, parameter, reference, internal-references, misc, and comment. 
;;; <li> The descriptions in .title, .precondition, .example, .comment, .return and .misc is arbitrary free text (without newlines).
;;; <li> There cannot be an internal tag on the first comment line. In other words, the first line must be a description.
;;; <li> Of parsing purposes, string quotes must be used around categories, anchors, refs and urls in .reference and .internal-references.
;;; </ol>
;;; The following internal tags are supported
;;; <pre>
;;;   .title Some title
;;;   .form (form-name par1 par2)
;;;   .pre-condition Some Precondition
;;;   .description Some Description
;;;   .parameter par-name parameter description
;;;   .example some example
;;;   .reference "category" "anchor" "url"
;;;   .internal-references "category" "ref1" "ref2" ...
;;;   .misc Miscelaneous information
;;;   .comment Internal comment
;;;   .returns Return value description
;;; </pre>
;;; 
;;; Internal references support categorized cross references within the current manual page. <p>
;;; 
;;; A reference is a categoried reference from an anchor to an arbitrary url.<p>
;;;
;;; We see that all manual clauses of the LAML manual document styles are supported except
;;; <kbd>parameters</kbd> (plural) and <kbd>cross-references</kbd>. These two are made
;;; automatically from a number of <kbd> parameter </kbd> clauses (singular), and a 
;;; number of <kbd> reference </kbd> and <kbd> internal-references </kbd> clauses (not necessarily adjecent).



; ---------------------------------------------------------------------------------------------------
; PART ONE:
; State machine operated traversal of input file, which turns lexical lisp comment to syntactical (comment ...) Lisp forms.
; Main function: (lexical-to-syntactical-comments! scheme-file)


;;; The first internal function. 
;;; The function in this section makes lexical Lisp style comments to syntactical comments.
;;; We document it because it may be useful in its own right.

(define state-list '()) ; for debugging purposes
(define debugging-lexical-to-syntactical-comments #f)


;; Given the source file, scheme-input-file, make another file scheme-output-file,
;; in which the lexical comments are turned into syntactic comments in terms of comment forms.
;; The first parameter of a comment form is the number of semicolons in the lexical comment.
;; The second parameter is the comment text itself. Several consequtive lines of (equally leveled
;; comments) are organized in one comment form.
(define (lexical-to-syntactical-comments! scheme-input-file scheme-output-file)
 (if (file-exists? scheme-output-file) (delete-file scheme-output-file))
 (let* ((ip (open-input-file scheme-input-file))
        (op (open-output-file scheme-output-file)))
     (lexical-to-syntactical-comments-given-ports ip op)
     (close-input-port ip)
     (close-output-port op)))


;; The name of the file where the function lexical-to-syntactical-comments! places it's output
(define temp-output-file "schemesource.tmp")


(define (lexical-to-syntactical-comments-given-ports inport outport)
  (set! state-list '())
  (lexical-to-syntactical-comments-given-ports-1 start-state 0 0 "" inport outport))

(define (lexical-to-syntactical-comments-given-ports-1 current-state semi-no new-semi-no remember-string inport outport)
  (if (not (eof-object? (peek-char inport)))
      (let* ((inch (as-string (read-char-drop-return inport)))
             (trans-res (lexical-to-syntactical-comments-transition current-state semi-no new-semi-no inch remember-string inport))
             (next-state (car trans-res))
             (next-semi-no (cadr trans-res))
             (next-new-semi-no (caddr trans-res))
             (toput (as-string (cadddr trans-res)))
             (next-remember-string (fifth trans-res))
            )
       (if debugging-lexical-to-syntactical-comments
            (set! state-list (cons trans-res state-list)))
       (write-string-to-port-scheme-doc toput outport)  
       (lexical-to-syntactical-comments-given-ports-1 next-state next-semi-no next-new-semi-no next-remember-string inport outport)
  )))

; Read and return the next character from inport.
; While the character is #\return, read on
(define (read-char-drop-return inport)
  (let ((ch (read-char inport)))
    (if (eq? ch #\return)
        (read-char-drop-return inport)
        ch)))

(define (eof-port? port) (eof-object? (peek-char port)))

(define (write-string-to-port-scheme-doc str port)
  (write-string-to-port-scheme-doc-1 str port 0 (string-length str)))

(define (write-string-to-port-scheme-doc-1 str port i str-length)
  (if (< i str-length)
      (begin 
        (write-char (string-ref str i) port)
        (write-string-to-port-scheme-doc-1 str port (+ i 1) str-length))))
  

; STATES 
; we maintain an additional-state, semi-no, which is the number of semicolons
; and also semi-no-new, which is the number of semicolons on the next comment line

; outside-comment                       within a portion of a file which is not a comment (and not within a string)
; outside-comment-within-string         as outside-comment, but within a string (in which ; characters do not signify comments)
; outside-comment-within-special-char   and outside-comment-within-special-char-1: State for reading special characters
; semicolon                             a semicolon just read
; within-comment                        inside a 'semi-no' semicolon comment
; leaving-comment                       inside a 'semi-no' semicolon comment, but CR just seen.
;                                       The comment may be continued on next line after blank space
; entering-comment                      We come from one comment and is now entering another. 
;                                       We don't know yet how many semicolons in the new
; left-comment                          we just left the comment.


(define start-state 'outside-comment)

(define NOT-RELEVANT 0)

(define (lexical-to-syntactical-comments-transition in-state semi-no semi-no-new inch remember-string inport)
 (let ()
   (cond 
         ((and (symbol? in-state) (eq? in-state 'outside-comment))
            (cond ((equal? inch ";")                                      (list  'semicolon 1 NOT-RELEVANT COMMENT-FORM-START ""))
                  ((equal? inch GATE-STRING)                              (list  'outside-comment-within-special-char 0 NOT-RELEVANT inch ""))
                  ((equal? inch QUOTE-STRING)                             (list  'outside-comment-within-string 0 NOT-RELEVANT inch ""))
                  (else                                                   (list  'outside-comment 0 NOT-RELEVANT inch ""))))

         ((and (symbol? in-state) (eq? in-state 'outside-comment-within-string))
            (cond ((equal? inch QUOTE-STRING)                             (list  'outside-comment 0 NOT-RELEVANT inch ""))
                  (else                                                   (list  'outside-comment-within-string 0 NOT-RELEVANT inch ""))))

         ((and (symbol? in-state) (eq? in-state 'semicolon))
            (cond ((equal? inch ";")                                      (list  'semicolon (+ 1 semi-no) NOT-RELEVANT EMPTY-STRING ""))
                  ((blank-or-tab? inch)                                   (list  'semicolon semi-no NOT-RELEVANT EMPTY-STRING "")) 
                  ((equal? inch CR-STRING)                                (list  'leaving-comment semi-no NOT-RELEVANT 
                                                                                  (string-append (as-string semi-no) SPACE-STRING QUOTE-STRING)    ; before 02.12.30: CR-STRING
                                                                                                 ""))      
                  ((equal? inch RETURN-STRING)                            (list  'semicolon semi-no NOT-RELEVANT EMPTY-STRING ""))   ; ignoring control M
                  ((equal? inch BACKSLASH-STRING)                         (list  'within-comment-backslash semi-no NOT-RELEVANT 
                                                                                  (string-append 
                                                                                   (as-string semi-no) SPACE-STRING QUOTE-STRING) ""))
                  ((equal? inch QUOTE-STRING)                             (list  'within-comment semi-no NOT-RELEVANT 
                                                                                 (string-append (as-string semi-no) SPACE-STRING QUOTE-STRING 
                                                                                                ESCAPED-QUOTE-STRING) ""))  ; more here
                  (else                                                   (list  'within-comment semi-no NOT-RELEVANT 
                                                                                 (string-append 
                                                                                   (as-string semi-no) SPACE-STRING QUOTE-STRING inch) ""))))

         ((and (symbol? in-state) (eq? in-state 'within-comment))
            (cond ((eof-port? inport)                                     (list  'left-comment NOT-RELEVANT  NOT-RELEVANT 
                                                                              (string-append  remember-string inch COMMENT-FORM-END CR-STRING) ""))
                  ((equal? inch CR-STRING)                                (list  'leaving-comment semi-no NOT-RELEVANT remember-string inch))
                  ((equal? inch RETURN-STRING)                            (list  'leaving-comment semi-no NOT-RELEVANT remember-string ""))   ; ignoring control M
                  ((equal? inch QUOTE-STRING)                             (list  'within-comment semi-no NOT-RELEVANT (string-append remember-string ESCAPED-QUOTE-STRING) ""))
                  ((equal? inch BACKSLASH-STRING)                         (list  'within-comment-backslash semi-no NOT-RELEVANT remember-string ""))
                  (else                                                   (list  'within-comment semi-no NOT-RELEVANT (string-append remember-string inch) ""))))

         ((and (symbol? in-state) (eq? in-state 'within-comment-backslash))
            (cond ((eof-port? inport)                                     (list  'left-comment NOT-RELEVANT  NOT-RELEVANT 
                                                                              (string-append  inch COMMENT-FORM-END CR-STRING) ""))
                  ((equal? inch CR-STRING)                                (list  'leaving-comment semi-no NOT-RELEVANT DOUBLE-BACKSLASH inch))
                  ((equal? inch RETURN-STRING)                            (list  'leaving-comment semi-no NOT-RELEVANT EMPTY-STRING ""))   ; ignoring control M
                  ((equal? inch QUOTE-STRING)                             (list  'within-comment semi-no NOT-RELEVANT ESCAPED-QUOTE-STRING ""))
                  ((equal? inch BACKSLASH-STRING)                         (list  'within-comment semi-no NOT-RELEVANT DOUBLE-BACKSLASH ""))
                  (else                                                   (list  'within-comment semi-no NOT-RELEVANT (string-append DOUBLE-BACKSLASH inch) ""))))

         ((and (symbol? in-state) (eq? in-state 'leaving-comment))
            (cond ((eof-port? inport)                                     (list  'left-comment NOT-RELEVANT  NOT-RELEVANT 
                                                                              (string-append  inch COMMENT-FORM-END CR-STRING) ""))
                  ((blank-space? inch)                                    (list  'leaving-comment semi-no NOT-RELEVANT "" 
                                                                                  (string-append remember-string inch)))
                  ((equal? inch ";")                                      (list  'entering-comment semi-no 1 remember-string ""))
                  ((equal? inch QUOTE-STRING)                             (list  'outside-comment-within-string 0 0 
                                                                                 (string-append COMMENT-FORM-END remember-string inch) "")) ; 30.12.99
                  (else                                                   (list  'left-comment 0 0 
                                                                                 (string-append COMMENT-FORM-END remember-string inch) "")
         )))

         ((and (symbol? in-state) (eq? in-state 'entering-comment))
            (cond ((equal? inch ";")                                      (list  'entering-comment semi-no (+ 1 semi-no-new) EMPTY-STRING ""))
                  ((equal? inch CR-STRING)                                (list  'leaving-comment semi-no NOT-RELEVANT CR-STRING ""))
                  ((blank-or-tab? inch)                                   (list  'within-comment semi-no-new NOT-RELEVANT
                                                                                 (if (= semi-no semi-no-new) 
                                                                                     EMPTY-STRING
                                                                                     (string-append 
                                                                                       COMMENT-FORM-END CR-STRING COMMENT-FORM-START
                                                                                        (string-append 
                                                                                         (as-string semi-no-new) SPACE-STRING QUOTE-STRING)))
                                                                                  ""
                                                                           ))
                  (else                                                   (list  'within-comment semi-no-new NOT-RELEVANT
                                                                                 (if (= semi-no semi-no-new) 
                                                                                     EMPTY-STRING
                                                                                     (string-append 
                                                                                       COMMENT-FORM-END CR-STRING COMMENT-FORM-START
                                                                                        (string-append 
                                                                                         (as-string semi-no-new) SPACE-STRING QUOTE-STRING)))
                                                                                  inch ; was ""
                                                                           ))))

         ((and (symbol? in-state) (eq? in-state 'left-comment))
            (cond ((equal? inch ";")                                      (list  'semicolon 1 NOT-RELEVANT COMMENT-FORM-START ""))
                  ((equal? inch QUOTE-STRING)                             (list  'outside-comment-within-string 0 NOT-RELEVANT inch "")); 30.12.99
                  (else                                                   (list  'outside-comment 0 NOT-RELEVANT inch ""))))

         ((and (symbol? in-state) (eq? in-state 'outside-comment-within-special-char))
            (cond ((equal? inch BACKSLASH-STRING)                         (list  'outside-comment-within-special-char-1 0 NOT-RELEVANT inch ""))
                  (else                                                   (list  'outside-comment 0 NOT-RELEVANT inch ""))))

         ((and (symbol? in-state) (eq? in-state 'outside-comment-within-special-char-1))  ; gate and backslash read: now reading last char
                                                                           (list  'outside-comment 0 NOT-RELEVANT inch ""))



         (else                                                            (error (string-append "lexical-to-syntactical-comments-transition error " 
                                                                                                (as-string in-state))))

  )))


(define (blank-space? ch-string)
  (or (equal? ch-string CR-STRING) (equal? ch-string TAB-STRING) (equal? ch-string SPACE-STRING)))

(define (blank-or-tab? ch-string)
  (or (equal? ch-string TAB-STRING) (equal? ch-string SPACE-STRING)))
    

; ---------------------------------------------------------------------------------------------------
; Special characters, represented as strings:

(define SPACE-STRING " ")
(define TAB-STRING (as-string #\tab))
(define QUOTE-STRING "\"")
(define ESCAPED-QUOTE-STRING (string-append (as-string #\\) "\""))
(define EMPTY-STRING "")
; (define CR-STRING "
; ")
(define CR-STRING (as-string (integer->char 10)))
(define RETURN-STRING (as-string (integer->char 13)))
(define GATE-STRING "#")
(define BACKSLASH-STRING (as-string #\\))
(define DOUBLE-BACKSLASH (string-append (as-string #\\) (as-string #\\)))

(define COMMENT-FORM-START (string-append "(comment "))
(define COMMENT-FORM-END (string-append QUOTE-STRING ")" CR-STRING))

; ---------------------------------------------------------------------------------------------------
; PART TWO:
; THE SYNTACTIC ANALYSIS OF THE TEMPORARY FILE WITH COMMENT FORMS:
; Main function: (extract-relevant-documentation file-name)

; State variable for the extraction of documentation:


;;; The second internal function. 
;;; The function in this section extracts relevant level n, n > 2, comments
;;; for documentation purposes


(define previous-level-2-comment "") ; always a string

; ---------------------------------------------------------------------------------------------------

; Return the contribution to the manual in terms of a manual entry (a list structure).
; return #f in case there is no contribution.
; This function reads exactly one form.
; In case a level 2 comment is read (at top level) remember it in the global variabel previous-level-2-comment and returns #f
(define (extract-next-documentation port)
  (if (not (eof-object? (peek-char port)))
      (let ((form-1 (read port)))
        (cond 
          ((comment-level? 4 form-1) 
                     (append-manual-abstract (contents-of-comment form-1)) 
                     (set! previous-level-2-comment "")
                     #f)

          ((comment-level? 3 form-1) 
                     (set! previous-level-2-comment "")
                     (make-manual-section-entry (contents-of-comment form-1)))
          
          ((comment-level? 2 form-1) 
                     (set! previous-level-2-comment (contents-of-comment form-1))
                     #f)

          ((define-form-scheme-doc? form-1) 
                     (let* ((signature (cadr form-1))
                            (item3 (caddr form-1))
                            (internal-level2-comment (if (comment-level? 2 item3) (contents-of-comment item3) #f))
                            (res (if (or internal-level2-comment (> (string-length previous-level-2-comment) 0))
                                     (make-manual-page-entry signature previous-level-2-comment internal-level2-comment)
                                     #f)))
                       (set! previous-level-2-comment "")
                       res))

          (else       (set! previous-level-2-comment "")
                      #f)))
      #f))

;; A variable holding the extracted manual abstract.
(define extracted-manual-abstract "")

(define (append-manual-abstract abstract)
  (set! extracted-manual-abstract (string-append extracted-manual-abstract " " abstract)))

; return whether form is a comment form at level n:
(define (comment-level? n form)
  (and (list? form) (> (length form) 1) (eq? (car form) 'comment) (number? (cadr form)) (= (cadr form) n)))

; return whether form is a define form:
(define (define-form-scheme-doc? form)
  (and (list? form) (> (length form) 1) (eq? (car form) 'define)))

; top level function in this section

;; Extract a manual style list of documentation from file-name and return it.
;; It is assumed that file-name contains syntactical (comment ...) comments.
(define (extract-relevant-documentation file-name)
 (let* ((ip (open-input-file file-name))
        (res (extract-relevant-documentation-from-port ip)))
     (close-input-port ip)
     res))

(define (extract-relevant-documentation-from-port port)
  (extract-relevant-documentation-from-port-iter port '()))

; ruturns the accumulated and manual formated doc-list when the temporary source file has been read
(define (extract-relevant-documentation-from-port-iter port doc-list)
  (if (not (eof-object? (peek-char port)))
      (let ((extract (extract-next-documentation port)))
	(if extract
	    (extract-relevant-documentation-from-port-iter port (cons extract doc-list))
	    (extract-relevant-documentation-from-port-iter port doc-list)))
      doc-list))

(define (make-manual-page-entry signature pre-comment inside-comment)
  ; pre-condition: signature and at least one of the comments are defined
  (let* ((kind "manual-page")
         (title (if (pair? signature) (as-string (car signature)) (as-string signature)))
         (form signature)
         (description (string-append
                        (if pre-comment  pre-comment "") CR-STRING
                        (if inside-comment  inside-comment "")))
         (parsed-description (extract-internal-tagging description))
        )

   (remove-duplicates-by-predicate
    (append
      (list (list 'kind kind))
      parsed-description
      (list 
        (list 'title title)
        (list 'form form)
      )
    )
    (lambda (x y)
      (and (eq? (car x) (car y))
           (not (memq (car x) allowed-duplicates-elements))))
   )
   

   ; in manual.scm: remove unknown tags.
   ; bundle parameter in parameters. Similar with references. 
   ; additional parsing of parameter etc.
  )
)


(define (make-manual-section-entry comment)
  (let ((kind "manual-section")
        (section-title (first-sentence-in-string comment))
        (section-body  (but-first-sentence-of-string comment)))
   (list 
     (list 'kind kind)
     (list 'section-title section-title)
     (list 'section-body section-body))))


(define (contents-of-comment c)
  (if (= 3 (length c))
      (caddr c)
      "??"))

; ---------------------------------------------------------------------------------------------------------------
;;; Internal tagging support in schemedoc comments.

; The character which designates internal tag start (on the start of a line).
(define internal-tagging-char #\.)

;; Parse the comment string, in order to extract internal tagging.
;; If no internal tagging is found return ((description "comment-string")).
;; If at least one internal tag is found return an a-list of the form
;; ((tag-1-symbol va1-1-string) (tag-2-symbol va1-2-string) ...). The string part
;; before the first tag is returned as (description "string part before first tag")
;; in the list.
(define (extract-internal-tagging comment-string)
 (let* ((lgt (string-length comment-string))
        (res-pos (first-internal-tag-position comment-string lgt 0))
       )
  (group-internal-tags
   (if res-pos
       (cons
         (list 'description (substring comment-string 0 (- res-pos 1)))
         (extract-internal-tagging-1 comment-string lgt (+ 1 res-pos)))
       (list (list 'description comment-string))))))

;; A list of element names (symbol) for which we allow multiple occurences.
(define allowed-duplicates-elements (list 'example))

; group the elements in itag-a-list into parameters, cross-references, and the others.
; Return an alist with parameters and cross-references clauses, in which parameter, reference, and internal-references 
; are embedded.
(define (group-internal-tags itag-a-list)
   (group-internal-tags-1 itag-a-list '() '() '()))

(define (group-internal-tags-1 itag-a-list pars refs others)
 (cond ((null? itag-a-list) 
          (append 
             (if (not (null? pars)) (list (cons 'parameters (reverse pars))) '())
             (if (not (null? refs)) (list (cons 'cross-references (reverse refs))) '())
             (reverse others)))
       (else
         (let ((itag (car itag-a-list)))
           (cond ((eq? 'parameter (car itag))
                    (group-internal-tags-1 (cdr itag-a-list) (cons (itag-parse-parameter (cadr itag)) pars) refs others))
                 ((eq? 'reference  (car itag))
                    (group-internal-tags-1 (cdr itag-a-list) pars  (cons (itag-parse-reference (cadr itag)) refs) others))
                 ((eq? 'internal-references  (car itag))
                    (group-internal-tags-1 (cdr itag-a-list) pars  (cons (itag-parse-internal-references (cadr itag)) refs) others))
                 (else (group-internal-tags-1 (cdr itag-a-list) pars refs (cons itag others)))))
       )))

; Return a parsed parameter clause, in which parameter name and explanation are sorted out.
(define (itag-parse-parameter par-desc-string)
  (let* ((lgt (string-length par-desc-string))
         (start-pos (skip-chars-in-string par-desc-string white-space-char-list 0))
         (space-after-name (find-in-string par-desc-string #\space start-pos))
         (tab-after-name (find-in-string par-desc-string #\tab start-pos))
         (pos (cond ((and space-after-name tab-after-name) (min space-after-name tab-after-name))
                    (space-after-name space-after-name)
                    (tab-after-name tab-after-name)
                    (else lgt)))
         (start-1-pos (skip-chars-in-string par-desc-string white-space-char-list pos))
        )
    (list 'parameter (substring par-desc-string start-pos pos) (substring par-desc-string start-1-pos lgt))))

; Return a parsed reference clause in which the three quoted constituents are sorted out.
(define (itag-parse-reference ref-desc-string)
  (let ((pair1 (find-delimited-string-in-string ref-desc-string #\" #\" 0)))
   (if pair1
       (let ((pair2 (find-delimited-string-in-string ref-desc-string #\" #\" (+ 1 (cdr pair1)))))
         (if pair2
             (let ((pair3 (find-delimited-string-in-string ref-desc-string #\" #\" (+ 1 (cdr pair2)))))
               (if pair3
                   (list 'reference 
                         (substring ref-desc-string (+ (car pair1) 1)  (cdr pair1))
                         (substring ref-desc-string (+ (car pair2) 1)  (cdr pair2))
                         (substring ref-desc-string (+ (car pair3) 1)  (cdr pair3)))
                   (list 'reference 
                         (substring ref-desc-string (+ (car pair1) 1)  (cdr pair1))
                         (substring ref-desc-string (+ (car pair2) 1)  (cdr pair2))
                         "???")))
             (list 'reference 
                    (substring ref-desc-string (+ (car pair1) 1) (cdr pair1))
                    "???"
                    "???")))
       (list 'reference "???" "???" "???"))))

; Return a list of at least length 2 in which the three quoted constituents are sorted out.
(define (itag-parse-internal-references int-ref-string)
 (let* ((res (itag-parse-internal-references-1 int-ref-string 0))
        (lgt (length res)))
   (cond ((null? res) (list 'internal-references "???" "???"))
         ((= 1 lgt)   (list 'internal-references (car res) "???"))
         (else        (cons 'internal-references res)))))

(define (itag-parse-internal-references-1 int-ref-string pos)
 (let ((pair (find-delimited-string-in-string int-ref-string #\" #\" pos)))
   (if pair
       (cons (substring int-ref-string (+ (car pair) 1) (cdr pair))
             (itag-parse-internal-references-1 int-ref-string (+ 1 (cdr pair))))
       '())))


; Return the indexes (cons pair) of the first substring of str delimited by
; ch1 and ch2 (in that order). (car result) designates ch1, and (cdr result) designates ch2.
; Return #f if not found. Start the search at start-pos
(define (find-delimited-string-in-string str ch1 ch2 start-pos)
 (let ((ch1-pos (find-in-string str ch1 start-pos)))
  (if ch1-pos
      (let ((ch2-pos (find-in-string str ch2 (+ 1 ch1-pos))))
        (if ch2-pos
            (cons ch1-pos ch2-pos)
            #f))
      #f)))
 

(define (first-internal-tag-position str lgt i)
  (let* ((cr-pos (find-in-string str #\newline i)))
   (if cr-pos
       (let ((after-blank-pos (skip-chars-in-string str white-space-char-list cr-pos)))         
         (cond ((not (number? after-blank-pos)) 
                   (error (string-append "first-internal-tag-position: Not number: " (as-string after-blank-pos))))
               ((and (< after-blank-pos lgt) (eq? internal-tagging-char (string-ref str after-blank-pos))) after-blank-pos)
               ((and (< after-blank-pos lgt) (not (eq? internal-tagging-char (string-ref str after-blank-pos))))
                   (first-internal-tag-position str lgt after-blank-pos))
               ((>= after-blank-pos lgt) #f)))
       #f)))
       
; an internal tag starts at tag-start-pos. tag-start-pos is an integer number.
; returns an a-list.
(define (extract-internal-tagging-1 comment-string lgt tag-start-pos)
 (if (not tag-start-pos) (error "extract-internal-tagging-1 error"))
 (let* ((space-pos (find-in-string comment-string #\space tag-start-pos))
        (eol-pos (find-in-string comment-string #\newline tag-start-pos))
        (next-tag-pos (if eol-pos (first-internal-tag-position comment-string lgt eol-pos) #f))
       )
   (cond ((and space-pos eol-pos (< space-pos eol-pos))
               (cons
                 (list (as-symbol (substring comment-string tag-start-pos space-pos)) (substring comment-string (+ 1 space-pos) eol-pos))
                 (if next-tag-pos (extract-internal-tagging-1 comment-string lgt (+ 1 next-tag-pos)) '())))
         ((and space-pos eol-pos (>= space-pos eol-pos))
                 (display-message (string-append "Dropping malformed internal tag:" (substring comment-string tag-start-pos eol-pos)))
                 (if next-tag-pos (extract-internal-tagging-1 comment-string lgt (+ 1 next-tag-pos)) '()))
         (space-pos 
               (cons
                 (list (as-symbol (substring comment-string tag-start-pos space-pos)) (substring comment-string (+ 1 space-pos) lgt))
                 '()))
         (eol-pos 
                 (display-message "Dropping malformed internal tag")
                 (if next-tag-pos (extract-internal-tagging-1 comment-string lgt (+ 1 next-tag-pos)) '()))
         (else '()))))


(define (well-formed? a-list)
  (if (list? a-list)
      (cond ((null? a-list) #t)
            ((and (list? (car a-list)) (= 2 (length (car a-list))) (symbol? (car (car a-list)))  (string? (cadr (car a-list))))
                #t)
            (else #f))
      #f)) 




